﻿Public Class Landscape

    Private heights(,) As Single                    ' výšky terénu v jednotlivých vrcholech
    Private verticesCount, indicesCount As Integer  ' počty vertexů a indexů
    Private width, height As Integer                ' šířka a výška oblasti

    Private device As GraphicsDevice                ' odkaz na objekt GraphicsDevice
    Private vb As VertexBuffer, ib As IndexBuffer   ' vertex a index buffer
    Private effect As BasicEffect                   ' výchozí efekt

    Private _texture As Texture2D                   ' textura terénu
    Public Property Texture() As Texture2D
        Get
            Return _texture
        End Get
        Set(ByVal value As Texture2D)
            _texture = value
        End Set
    End Property

    Private _textureScale As Single = 1.0F
    Public Property TextureScale() As Single
        Get
            Return _textureScale
        End Get
        Set(ByVal value As Single)
            _textureScale = value
        End Set
    End Property



    Public Sub New(ByVal graphics As GraphicsDevice)
        'schovat si objekt GraphicsDevice
        Me.device = graphics

        'připravit efekt
        effect = New BasicEffect(device, Nothing)
    End Sub

    Public Sub LoadTerrain(ByVal tex As Texture2D, ByVal heightScale As Double)
        ' zjistit rozměry terénu
        Me.Width = tex.Width : Me.Height = tex.Height
        ReDim heights(Me.Width - 1, Me.Height - 1)

        ' získat jednotlivé pixely textury
        Dim b(Width * Height - 1) As Color
        tex.GetData(Of Color)(b)

        ' zjistit výšky v jednotlivých bodech
        For y As Integer = 0 To Height - 1
            For x As Integer = 0 To Width - 1
                With b(x + y * Width)
                    heights(x, y) = heightScale / 255 * (CDbl(.R) + CDbl(.G) + CDbl(.B)) / 3
                End With
            Next
        Next

        ' připravit vertex a index buffer
        PrepareVertexBuffer()
        PrepareIndexBuffer()
    End Sub

    Private Sub PrepareVertexBuffer()
        ' nastavit vertex buffer
        verticesCount = width * height
        vb = New VertexBuffer(device, GetType(VertexPositionNormalTexture), verticesCount, BufferUsage.WriteOnly)

        ' vytvořit pole vertexů
        Dim verts(verticesCount - 1) As VertexPositionNormalTexture
        For y As Integer = 0 To height - 1
            For x As Integer = 0 To width - 1
                With verts(x + y * width)
                    .Position = New Vector3(x, heights(x, y), y)

                    .TextureCoordinate = New Vector2(x * _textureScale / width, y * _textureScale / height)

                    If x > 0 And x < width - 1 And y > 0 And y < height - 1 Then
                        Dim v1 As New Vector3(2, heights(x + 1, y) - heights(x - 1, y), 0)
                        Dim v2 As New Vector3(0, heights(x, y + 1) - heights(x, y - 1), 2)
                        .Normal = Vector3.Cross(v2, v1)
                        .Normal.Normalize()
                    Else
                        .Normal = Vector3.Up
                    End If
                End With
            Next
        Next

        ' nahrát vertexy do bufferu
        vb.SetData(Of VertexPositionNormalTexture)(verts)
    End Sub

    Private Sub PrepareIndexBuffer()
        ' nastavit index buffer
        indicesCount = (width - 1) * (height - 1) * 6
        ib = New IndexBuffer(device, GetType(Integer), indicesCount, BufferUsage.WriteOnly)

        ' vytvořit pole indexů
        Dim indices(indicesCount - 1) As Integer
        Dim k As Integer = 0
        For y As Integer = 0 To height - 2
            For x As Integer = 0 To width - 2
                Dim b As Integer = x + y * width
                indices(k) = b
                indices(k + 1) = b + 1
                indices(k + 2) = b + 1 + width
                indices(k + 3) = b
                indices(k + 4) = b + 1 + width
                indices(k + 5) = b + width
                k += 6
            Next
        Next

        ' nahrát indexy do bufferu
        ib.SetData(Of Integer)(indices)
    End Sub

    Public Sub Draw(ByVal world As Matrix, ByVal view As Matrix, ByVal proj As Matrix)
        ' nastavit projekční matice
        effect.World = world
        effect.View = view
        effect.Projection = proj

        ' zapnout texturování
        effect.Texture = Texture
        effect.TextureEnabled = True

        ' osvětlení scény
        effect.LightingEnabled = True
        effect.DirectionalLight0.Enabled = True
        effect.DirectionalLight0.Direction = New Vector3(15, -7, -18)           ' směr světla
        effect.DirectionalLight0.DiffuseColor = New Vector3(1, 1, 0.7)          ' barva světla
        effect.DirectionalLight0.SpecularColor = New Vector3(0.5, 0.5, 0.32)    ' nasvícené plochy
        effect.AmbientLightColor = New Vector3(0.7, 0.7, 0.7)                   ' stín (tam, kam nejde světlo)

        ' vykreslit jen mřížku
        device.RenderState.FillMode = FillMode.Solid

        ' použít efekt na vykreslování
        effect.Begin(SaveStateMode.None)
        For Each pass As EffectPass In effect.CurrentTechnique.Passes       ' všechny průchody
            pass.Begin()

            ' nastavit formát vertexu
            device.VertexDeclaration = New VertexDeclaration(device, VertexPositionNormalTexture.VertexElements)
            ' nastavit vertex a index buffer
            device.Vertices(0).SetSource(vb, 0, VertexPositionNormalTexture.SizeInBytes)
            device.Indices = ib
            ' vykreslit vertexy
            device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, verticesCount, 0, indicesCount / 3)

            pass.End()
        Next
        effect.End()
    End Sub

    Private Function Interpolate(ByVal val1 As Single, ByVal val2 As Single, ByVal koef As Single) As Single
        'provede lineární interpolaci a vrátí výsledek
        '- val1 a val2 jsou hodnoty v krejních bodech
        '- koef je číslo od 0 do 1 reprezentující vzdálenost od prvního bodu
        Return val1 + (val2 - val1) * koef
    End Function

    Public Function GetHeight(ByVal x As Single, ByVal z As Single) As Single
        'pokud jsme náhodou mimo terén, vrátit nulovou výšku
        If x < 0.0F Or z < 0.0F Or x >= width - 1 Or z >= height - 1 Then Return 0.0F

        'ošetřit body ve skoro celočíselných vrcholech
        If Math.Abs(x - Math.Round(x)) < 0.001F And Math.Abs(z - Math.Round(z)) < 0.001F Then Return heights(Math.Round(x), Math.Round(z))

        'zjistit souřadnice horního levého rohu čtverce
        Dim lX As Integer = Math.Floor(x), lZ As Integer = Math.Floor(z)

        'určit, ve kterém trojúhelníku jsme
        Dim rX As Single = x - lX, rZ As Single = z - lZ
        If rX > rZ Then          'pravý trojúhelník
            Dim bodZ As Single = Interpolate(lZ, z, 1.0F / rX)   'spočítat bod.Z
            Dim bodY As Single = Interpolate(heights(lX + 1, lZ), heights(lX + 1, lZ + 1), bodZ - lZ)   'spočítat bod.Y
            Dim koef As Single = Math.Sqrt(rX * rX + rZ * rZ) / Math.Sqrt(1 * 1 + (bodZ - lZ) * (bodZ - lZ))   'spočítat poměr vzdáleností
            Return Interpolate(heights(lX, lZ), bodY, koef)

        Else                             'levý trojúhelník
            Dim bodX As Single = Interpolate(lX, x, 1.0F / rZ)   'spočítat bod.Z
            Dim bodY As Single = Interpolate(heights(lX, lZ + 1), heights(lX + 1, lZ + 1), bodX - lX)   'spočítat bod.Y
            Dim koef As Single = Math.Sqrt(rX * rX + rZ * rZ) / Math.Sqrt((bodX - lX) * (bodX - lX) + 1 * 1)   'spočítat poměr vzdáleností
            Return Interpolate(heights(lX, lZ), bodY, koef)

        End If
    End Function

End Class
